#include <AvailabilityMacros.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/mman.h>
#include <inttypes.h>
#include <IOKit/IOKitLib.h>
#include <CoreFoundation/CoreFoundation.h>

#include "mapMemoryDriverShared.h"

void printArray( unsigned long offset, UInt8* array, unsigned long len ) {
	if (!len) return;
	unsigned int i;
	printf("0x%08lx : ", offset);
	for (i=0; i<len; i++) {
		printf("%02x ",array[i]);
	}
	printf("\n");
}

int main( int argc, char * argv[] )
{
    io_iterator_t iter;
    io_service_t service, parent;
	io_connect_t connect;
	io_name_t dev_name;
    kern_return_t kr;
	UInt16 device_id=0, vendor_id=0;
	CFDataRef data_ref;	
	CFMutableDictionaryRef properties;
	
#if __LP64__
    mach_vm_address_t		addr;
    mach_vm_size_t			size;
#else
    vm_address_t			addr;
    vm_size_t				size;
#endif
	
	int c=0, i=0, target=0;
	unsigned long offset=0, max_size=0x1000, payload_size=128;
	bool driverFound = false, base16 = false, file_dump = false, screen_dump = false;
	
	while ((c = getopt (argc, argv, "Sdn:o:s:p:h?")) != -1)
		switch (c)
	{
		case 'S':
			screen_dump = true;
			payload_size = 32;
			break;
		case 'd':
			file_dump = true;
			break;
		case 'n':
			target = atoi(optarg);
			break;
		case 'o':
			base16 = optarg[0]=='0' && (optarg[1]=='x' || optarg[1]=='X');
			offset = strtol(optarg,NULL,(base16?16:10));
			break;
		case 's':
			base16 = optarg[0]=='0' && (optarg[1]=='x' || optarg[1]=='X');
			max_size = strtol(optarg,NULL,(base16?16:10));
			break;
		case 'p':
			base16 = optarg[0]=='0' && (optarg[1]=='x' || optarg[1]=='X');
			payload_size = strtol(optarg,NULL,(base16?16:10));
			break;
		case '?':
		case 'h':
		default:
			printf("\nUsage: %s [-S] [-d] [-n num] [-o offset] [-s max_size] [-p payload_size]\n\n",argv[0]);
			printf("  [ -S ]              Print to screen\n");
			printf("  [ -d ]              Dump to file\n");
			printf("  [ -n num ]          Driver number\n");
			printf("  [ -o offset ]       Start offset\n");
			printf("  [ -s max_size ]     Max output length\n");
			printf("  [ -p payload_size ] Chunking size\n");
			printf("\n");
			exit(1);
			break;
	}			
	
	bool dump = screen_dump || file_dump;
	UInt8 payload[(screen_dump?payload_size:0)];
	
    kr = IOServiceGetMatchingServices( kIOMasterPortDefault,
									  IOServiceMatching( kMapMemoryDriver ), &iter);
    assert( KERN_SUCCESS == kr );
	
	if (dump) printf("      Offset = 0x%08lx (%ld)\n", offset, offset);
	if (dump) printf("    Max Size = 0x%08lx (%ld)\n", max_size, max_size);
	if (dump) printf("Payload Size = %ld\n\n", payload_size);
	
    for( ;
		(service = IOIteratorNext(iter));
		IOObjectRelease(service)) {
		
		driverFound = true;
		i++;
		
        io_string_t path;
		
		kr = IORegistryEntryGetParentEntry( service, kIOServicePlane, &parent );
		assert( KERN_SUCCESS == kr );
		
        kr = IORegistryEntryGetPath(parent, kIOServicePlane, path);
        assert( KERN_SUCCESS == kr );
		
        if (!dump) {
			printf("%d : %s\n", i, path);
			continue;
		}		
		if (target && target!=i) continue; 
		
		kr = IORegistryEntryCreateCFProperties( parent, &properties, kCFAllocatorDefault, kNilOptions );
		assert( KERN_SUCCESS == kr );
		
		kr = IORegistryEntryGetName(parent, dev_name);
		assert( KERN_SUCCESS == kr );
		
		data_ref = (CFDataRef)CFDictionaryGetValue(properties, CFSTR("vendor-id"));
		if (data_ref) vendor_id = *(UInt16*)CFDataGetBytePtr(data_ref);
		data_ref = (CFDataRef)CFDictionaryGetValue(properties, CFSTR("device-id"));
		if (data_ref) device_id = *(UInt16*)CFDataGetBytePtr(data_ref);
		
		kr = IOServiceOpen( service, mach_task_self(), 0, &connect );
		assert( KERN_SUCCESS == kr );
		
		char f[1024];
		snprintf(f,1024,"%d-%s-%04x-%04x.bin",i,dev_name,vendor_id,device_id);
		printf("Output: %s\n",f);
		
		kr = IOConnectMapMemory( connect, 0,
								mach_task_self(), &addr, &size,
								kIOMapAnywhere | kIOMapDefaultCache );
		if (KERN_SUCCESS != kr) {
			printf("* Connect Map Memory failed! (root required?)\n\n");
			continue;
		}		
		
		if (offset>size) {
			printf("* Skipping, offset too big.\n\n");
			continue;
		}
		
		unsigned long my_size = (max_size && max_size<(size-offset) ? max_size : (size-offset));
		FILE* file = NULL;
		int filedes;
		char* fmap = NULL;
		
		if (file_dump) {
			file = fopen(f,"w+");
			assert(file);
			filedes = fileno(file);
			ftruncate(filedes,my_size);
			
			fmap = mmap(0,my_size,PROT_WRITE,MAP_SHARED,filedes,0);
			if(fmap==(char*)-1){
				printf("Got mmap errno %d\n",errno);
				exit(1);
			}
		}
		
		unsigned long curr_off;
		for (curr_off=0; payload_size<=my_size && curr_off<(my_size-payload_size+1); curr_off+=payload_size) {
			if (screen_dump) {
				memcpy(payload,(void*)(addr+offset+curr_off),payload_size);
				printArray(offset+curr_off,payload,payload_size);
			}
			if (file_dump) {
				memcpy(fmap+curr_off,(void*)(addr+offset+curr_off),payload_size);
			}
		}
		if (screen_dump) {
			memcpy(payload,(void*)(addr+offset+curr_off),my_size%payload_size);
			printArray(offset+curr_off,payload,my_size%payload_size);			
		}
		if (file_dump) {
			memcpy(fmap+curr_off,(void*)(addr+offset+curr_off),my_size%payload_size);
			munmap(fmap,my_size);
			fclose(file);
		}
		
		kr = IOConnectUnmapMemory( connect, 0, mach_task_self(), addr );
		assert( KERN_SUCCESS == kr );
	}
	
    IOObjectRelease(iter);
	
	if (driverFound == false) {
		fprintf(stderr, "No matching drivers found.\n");
	}
	
    return EXIT_SUCCESS;
}

